package auth

import (
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net/http"
	"strings"

	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

type BearerAuthHeaderBuilder struct {
	uaaURL       string
	clientID     string
	clientSecret string

	// only overridden in system test code, to add a custom root CA for BOSH
	HttpClient boshclient.HTTPClient
}

func NewBearerAuthHeaderBuilder(uaaURL, clientID, clientSecret string) *BearerAuthHeaderBuilder {
	return &BearerAuthHeaderBuilder{
		uaaURL:       uaaURL,
		clientID:     clientID,
		clientSecret: clientSecret,
		HttpClient:   http.DefaultClient,
	}
}

func (hb *BearerAuthHeaderBuilder) Build() (string, error) {
	request, err := http.NewRequest(http.MethodPost, fmt.Sprintf("%s/oauth/token", hb.uaaURL), strings.NewReader("grant_type=client_credentials"))
	if err != nil {
		return "", err
	}
	request.Header.Set("Content-type", "application/x-www-form-urlencoded")
	request.SetBasicAuth(hb.clientID, hb.clientSecret)

	response, err := hb.HttpClient.Do(request)
	if err != nil {
		return "", err
	}
	defer response.Body.Close()

	if response.StatusCode != http.StatusOK {
		bodyBytes, err := ioutil.ReadAll(response.Body)
		if err != nil {
			bodyBytes = []byte("<error reading body>")
		}
		return "", fmt.Errorf("Error authenticating (%d): %s", response.StatusCode, bodyBytes)
	}

	responseContent := map[string]interface{}{}
	err = json.NewDecoder(response.Body).Decode(&responseContent)
	if err != nil {
		return "", err
	}

	accessToken, ok := responseContent["access_token"]
	if !ok {
		return "", fmt.Errorf("no access token in grant %#v", responseContent)
	}

	return fmt.Sprintf("Bearer %s", accessToken), nil
}
