package boshclient

import (
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"strconv"
	"strings"
	"time"

	"github.com/craigfurman/herottp"
)

type Client struct {
	boshURL string

	BoshPollingInterval time.Duration

	authHeaderBuilder AuthHeaderBuilder
	HTTPClient        HTTPClient
	logger            *log.Logger
}

type HTTPClient interface {
	Do(req *http.Request) (*http.Response, error)
}

//go:generate counterfeiter -o fake_auth_header_builder/fake_auth_header_builder.go . AuthHeaderBuilder
type AuthHeaderBuilder interface {
	Build() (string, error)
}

func New(boshURL string, authHeaderBuilder AuthHeaderBuilder, logger *log.Logger) *Client {
	return &Client{
		boshURL: boshURL,

		authHeaderBuilder: authHeaderBuilder,
		HTTPClient: herottp.New(herottp.Config{
			NoFollowRedirect: true,
		}),
		logger:              logger,
		BoshPollingInterval: 5,
	}
}

type BoshTask struct {
	State       string
	Description string
	Result      string
}

const (
	Queued     = "queued"
	Processing = "processing"
	Done       = "done"
	Error      = "error"

	BoshTaskCreateDeployment       = "create deployment"
	BoshTaskDeleteDeploymentPrefix = "delete deployment"
)

func taskIDFromRedirect(response *http.Response) (int, error) {
	if response.StatusCode != http.StatusFound {
		defer response.Body.Close()
		responseBytes, _ := ioutil.ReadAll(response.Body)
		return 0, fmt.Errorf("expected status 302, got %d. Body: %s", response.StatusCode, string(responseBytes))
	}

	urlParts := strings.Split(response.Header.Get("location"), "/")
	taskID, err := strconv.Atoi(urlParts[len(urlParts)-1])
	if err != nil {
		return 0, fmt.Errorf("error parsing task ID: %s", err)
	}

	return taskID, nil
}

func unexpectedStatusErr(response *http.Response, expectedStatus int) error {
	defer response.Body.Close()
	body, err := ioutil.ReadAll(response.Body)
	var bodyStr string
	if err == nil {
		bodyStr = string(body)
	} else {
		bodyStr = "COULDN'T READ RESPONSE BODY"
	}

	return fmt.Errorf("expected status %d, was %d. Response Body: %s", expectedStatus, response.StatusCode, bodyStr)
}
