package boshclient_test

import (
	"errors"
	"fmt"

	"github.com/pivotal-cf/on-demand-service-broker/boshclient"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

func itBuildsAuthorizationHeader(err func() error) {
	It("calls the authorization header builder", func() {
		Expect(authHeaderBuilder.BuildCallCount()).To(BeNumerically(">", 0))
	})

	Context("when the Authorization header cannot be generated", func() {
		BeforeEach(func() {
			authHeaderBuilder.BuildReturns("", errors.New("some-error"))
		})

		It("returns an error", func() {
			Expect(err()).To(MatchError(ContainSubstring("some-error")))
		})
	})
}

var _ = Describe("bosh client", func() {

	Describe("deleting bosh deployments", func() {
		var (
			taskID    int
			deleteErr error
		)

		JustBeforeEach(func() {
			director.InitializeDeleteDeploymentTest(shouldFail, 90)
			taskID, deleteErr = c.DeleteDeployment("some-deployment")
		})

		itBuildsAuthorizationHeader(func() error { return deleteErr })

		Context("when bosh can delete the deployment", func() {
			BeforeEach(func() {
				shouldFail = false
			})

			It("calls bosh to delete a deployment by name", func() {
				Expect(director.DeleteDeploymentCalls).To(Equal(1))
				Expect(director.DeletedDeploymentID).To(Equal("some-deployment"))
			})

			It("returns the bosh task ID", func() {
				Expect(taskID).To(Equal(90))
			})
		})

		Context("when bosh cannot delete the deployment", func() {
			BeforeEach(func() {
				shouldFail = true
			})

			It("returns an error", func() {
				Expect(deleteErr).To(MatchError(ContainSubstring("expected status 302, got 500")))
			})
		})
	})

	Describe("getting VM info", func() {
		var (
			name = "some-deployment"

			vms    boshclient.BoshVMs
			vmsErr error

			taskIDToReturn = 42
		)

		BeforeEach(func() {
			director.InitializeVMsTest(taskIDToReturn, false)
			director.InitializeTaskTest(taskIDToReturn, boshclient.BoshTask{State: boshclient.Done}, false)
			director.InitializeVMsOutputTest(taskIDToReturn, []boshclient.BoshVMsOutput{{IPs: []string{"ip1", "ip2"}, InstanceGroup: "an-instance-group"}}, false)
		})

		JustBeforeEach(func() {
			vms, vmsErr = c.VMs(name)
		})

		itBuildsAuthorizationHeader(func() error { return vmsErr })

		It("queries BOSH for VMs info for deployment with given name", func() {
			Expect(director.VMsCalls).To(Equal(1))
			Expect(director.VMsDeploymentName).To(Equal(name))
		})

		Context("when bosh starts VMs task successfully", func() {
			Context("when task state can be fetched from bosh successfully", func() {
				Context("when task output can be fetched from bosh successfully", func() {
					It("returns BOSH VMs", func() {
						Expect(vms).To(HaveLen(1))
						Expect(vms["an-instance-group"]).To(ConsistOf("ip1", "ip2"))
					})

					It("returns no error", func() {
						Expect(vmsErr).NotTo(HaveOccurred())
					})
				})

				Context("when the task is finished, but has failed", func() {
					BeforeEach(func() {
						director.InitializeTaskTest(taskIDToReturn, boshclient.BoshTask{State: boshclient.Error}, false)
					})

					It("returns an error", func() {
						Expect(vmsErr).To(MatchError(ContainSubstring(fmt.Sprintf("task %d failed", taskIDToReturn))))
					})
				})

				Context("when fetching task output from bosh fails", func() {
					BeforeEach(func() {
						director.InitializeVMsOutputTest(taskIDToReturn, nil, true)
					})

					It("returns an error", func() {
						Expect(vmsErr).To(MatchError(ContainSubstring("expected status 302, was 500")))
					})
				})
			})

			Context("when fetching task state from bosh fails", func() {
				BeforeEach(func() {
					director.InitializeTaskTest(taskIDToReturn, boshclient.BoshTask{}, true)
				})

				It("returns an error", func() {
					Expect(vmsErr).To(MatchError(ContainSubstring("expected status 302, was 500")))
				})
			})
		})

		Context("when bosh fails to start VMs task", func() {
			BeforeEach(func() {
				director.InitializeVMsTest(taskIDToReturn, true)
			})

			It("returns an error", func() {
				Expect(vmsErr).To(MatchError(ContainSubstring("expected status 302, was 500")))
			})
		})
	})

	// Leaving these tests here for now, but output method will need to be made more generic at some point,
	// e.g. getting output from a failed deploy task to log it / expose it over an API.
	// At the moment, output is only used in the VMs call (above)
	Describe("getting output from a BOSH VMs task", func() {
		var (
			taskID            = 2
			expectedVMsOutput = []boshclient.BoshVMsOutput{{IPs: []string{"a-nice-ip"}, InstanceGroup: "a-nice-instance-group"}}

			boshOutput []boshclient.BoshVMsOutput
			outputErr  error
		)

		JustBeforeEach(func() {
			boshOutput, outputErr = c.VMsOutput(taskID)
		})

		Context("when bosh succeeds", func() {
			BeforeEach(func() {
				director.InitializeVMsOutputTest(taskID, expectedVMsOutput, false)
			})

			itBuildsAuthorizationHeader(func() error { return outputErr })

			It("does not error", func() {
				Expect(outputErr).NotTo(HaveOccurred())
			})

			It("returns the IPs in the output", func() {
				Expect(boshOutput).To(ConsistOf(boshclient.BoshVMsOutput{IPs: []string{"a-nice-ip"}, InstanceGroup: "a-nice-instance-group"}))
			})
		})

		Context("when bosh fails", func() {
			BeforeEach(func() {
				director.InitializeVMsOutputTest(taskID, nil, true)
			})

			It("returns an error", func() {
				Expect(outputErr).To(MatchError(ContainSubstring("expected status 200, was 500.")))
			})
		})
	})

	Describe("downloading deployment manifest, as a byte array", func() {
		var (
			name             string
			manifest         []byte
			manifestFetchErr error
		)

		JustBeforeEach(func() {
			manifest, manifestFetchErr = c.GetManifestRaw(name)
		})

		Context("when the deployment exists", func() {
			BeforeEach(func() {
				name = "a-manifest"
				director.InitializeGetManifestTest([]byte("a-manifest"), false)
			})

			itBuildsAuthorizationHeader(func() error { return manifestFetchErr })

			It("returns the manifest", func() {
				Expect(manifest).To(Equal([]byte("a-manifest")))
			})

			It("returns no error", func() {
				Expect(manifestFetchErr).NotTo(HaveOccurred())
			})
		})

		Context("when the deployment does not exist", func() {
			BeforeEach(func() {
				name = "a-manifest"
				director.InitializeGetManifestTest(nil, false)
			})

			It("returns the manifest", func() {
				Expect(manifest).To(BeNil())
			})

			It("returns no error", func() {
				Expect(manifestFetchErr).NotTo(HaveOccurred())
			})
		})
	})
	Describe("downloading deployment manifest", func() {
		var (
			name             string
			manifest         *boshclient.BoshManifest
			manifestFetchErr error
		)

		JustBeforeEach(func() {
			manifest, manifestFetchErr = c.GetManifest(name)
		})

		Context("when the deployment exists", func() {
			BeforeEach(func() {
				name = "a-manifest"
				director.InitializeGetManifestTest([]byte(`name: a-manifest
director_uuid: ""
releases: []
stemcells: []
instance_groups: []
update:
  canaries: 0
  canary_watch_time: ""
  update_watch_time: ""
  max_in_flight: 0
`), false)
			})

			itBuildsAuthorizationHeader(func() error { return manifestFetchErr })

			It("returns the manifest", func() {
				Expect(manifest.Name).To(Equal(name))
			})

			It("returns no error", func() {
				Expect(manifestFetchErr).NotTo(HaveOccurred())
			})
		})

		Context("when the deployment does not exist", func() {
			BeforeEach(func() {
				name = "a-manifest"
				director.InitializeGetManifestTest(nil, false)
			})

			It("returns the manifest", func() {
				Expect(manifest).To(BeNil())
			})

			It("returns no error", func() {
				Expect(manifestFetchErr).NotTo(HaveOccurred())
			})
		})

		Context("when the manifest cannot be fetched from BOSH", func() {
			BeforeEach(func() {
				name = "a-manifest"
				director.InitializeGetManifestTest(nil, true)
			})

			It("returns an error", func() {
				Expect(manifestFetchErr).To(MatchError(ContainSubstring("expected status 200")))
			})
		})
	})
})
