package boshclient_test

import (
	"io/ioutil"
	"os"
	"path/filepath"

	"github.com/pivotal-cf/on-demand-service-broker/boshclient"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"gopkg.in/yaml.v2"
)

var _ = Describe("de(serialising) BOSH manifests", func() {
	sampleManifest := boshclient.BoshManifest{
		Name:         "deployment-name",
		DirectorUUID: "76f2d2b1-fc3f-47d2-bbfe-4109dc799337",
		Releases: []boshclient.Release{
			{
				Name:    "a-release",
				Version: "latest",
			},
		},
		Stemcells: []boshclient.Stemcell{
			{
				Alias:   "greatest",
				OS:      "Windows",
				Version: "3.1",
			},
		},
		InstanceGroups: []boshclient.InstanceGroup{
			{
				Name:      "jerb",
				Instances: 1,
				Jobs: []boshclient.Job{
					{
						Name:    "broker",
						Release: "a-release",
						Provides: map[string]boshclient.ProvidesLink{
							"some_link": {As: "link-name"},
						},
						Consumes: map[string]boshclient.ConsumesLink{
							"another_link": {From: "jerb-link"},
						},
					},
				},
				VMType:             "massive",
				PersistentDiskType: "big",
				AZs:                []string{"az1", "az2"},
				Stemcell:           "greatest",
				Networks: []boshclient.Network{
					{
						Name:      "a-network",
						StaticIPs: []string{"10.0.0.0"},
					},
				},
			},
			{
				Name:      "an-errand",
				Lifecycle: "errand",
				Instances: 1,
				Jobs: []boshclient.Job{
					{
						Name:    "a-release",
						Release: "a-release",
					},
				},
				VMType:   "small",
				Stemcell: "greatest",
				Networks: []boshclient.Network{
					{
						Name: "a-network",
					},
				},
			},
		},
		Properties: map[string]interface{}{
			"foo": "bar",
		},
		Update: boshclient.Update{
			Canaries:        1,
			CanaryWatchTime: "30000-180000",
			UpdateWatchTime: "30000-180000",
			MaxInFlight:     4,
		},
	}

	It("deserialises BOSH manifests", func() {
		cwd, err := os.Getwd()
		Expect(err).NotTo(HaveOccurred())
		manifestBytes, err := ioutil.ReadFile(filepath.Join(cwd, "fixtures", "manifest.yml"))
		Expect(err).NotTo(HaveOccurred())
		var manifest boshclient.BoshManifest
		Expect(yaml.Unmarshal(manifestBytes, &manifest)).To(Succeed())

		Expect(manifest).To(Equal(sampleManifest))
	})

	It("serialises and deserialises a bosh manifest", func() {
		content, err := yaml.Marshal(sampleManifest)
		Expect(err).NotTo(HaveOccurred())
		var actualManifest boshclient.BoshManifest
		Expect(yaml.Unmarshal(content, &actualManifest)).To(Succeed())
		Expect(actualManifest).To(Equal(sampleManifest))
	})

	It("omits optional keys", func() {
		emptyManifest := boshclient.BoshManifest{
			Releases: []boshclient.Release{
				{},
			},
			Stemcells: []boshclient.Stemcell{
				{},
			},
			InstanceGroups: []boshclient.InstanceGroup{
				{
					Networks: []boshclient.Network{
						{},
					},
				},
			},
		}

		content, err := yaml.Marshal(emptyManifest)
		Expect(err).NotTo(HaveOccurred())

		Expect(content).NotTo(ContainSubstring("static_ips:"))
		Expect(content).NotTo(ContainSubstring("lifecycle:"))
		Expect(content).NotTo(ContainSubstring("azs:"))
		Expect(content).NotTo(ContainSubstring("persistent_disk_type:"))
		Expect(content).NotTo(ContainSubstring("jobs:"))
		Expect(content).NotTo(ContainSubstring("provides:"))
		Expect(content).NotTo(ContainSubstring("consumes:"))
	})
})
