package boshclient_test

import (
	"log"

	"gopkg.in/yaml.v2"

	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient/auth"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("deploying a bosh release", func() {
	const (
		taskID = 3
	)

	var (
		deploymentManifest = boshclient.BoshManifest{
			Name:           "big deployment",
			Releases:       []boshclient.Release{},
			Stemcells:      []boshclient.Stemcell{},
			InstanceGroups: []boshclient.InstanceGroup{},
		}
		deploymentManifestBytes []byte

		returnedTaskID int
		deployErr      error
	)

	BeforeEach(func() {
		shouldFail = false
		var err error
		deploymentManifestBytes, err = yaml.Marshal(deploymentManifest)
		Expect(err).NotTo(HaveOccurred())
	})

	JustBeforeEach(func() {
		director.InitializeDeployTest(taskID, shouldFail)
	})

	AfterEach(func() {
		director.Close()
	})

	Context("using a structured manifest", func() {
		JustBeforeEach(func() {
			returnedTaskID, deployErr = c.Deploy(deploymentManifest)
		})

		itBuildsAuthorizationHeader(func() error { return deployErr })

		Context("when bosh deploys the release successfully", func() {
			It("deploys release with specified name", func() {
				Expect(deployErr).NotTo(HaveOccurred())
				Expect(director.DeployCalls()).To(Equal(1))
				Expect(director.ManifestBytes).To(Equal(deploymentManifestBytes))
			})

			It("returns the task ID", func() {
				Expect(returnedTaskID).To(Equal(taskID))
			})
		})

		Context("when BOSH cannot deploy the release", func() {
			BeforeEach(func() {
				shouldFail = true
			})

			It("returns error", func() {
				Expect(deployErr).To(MatchError(ContainSubstring("expected status 302, was 500.")))
			})
		})

		Context("when the BOSH director cannot be reached", func() {
			It("returns error", func() {
				c = boshclient.New("http://localhost", auth.NewBasicAuthHeaderBuilder("", ""), log.New(GinkgoWriter, "[broker-unit-tests] ", log.LstdFlags))
				_, err := c.Deploy(deploymentManifest)
				Expect(err).To(MatchError(ContainSubstring("error contacting BOSH director")))
			})
		})
	})

	Context("using a raw manifest", func() {
		JustBeforeEach(func() {
			deploymentManifestBytes, err := yaml.Marshal(deploymentManifest)
			Expect(err).NotTo(HaveOccurred())
			returnedTaskID, deployErr = c.DeployRaw(deploymentManifestBytes)
		})

		itBuildsAuthorizationHeader(func() error { return deployErr })

		Context("when bosh deploys the release successfully", func() {
			It("deploys release with specified name", func() {
				Expect(deployErr).NotTo(HaveOccurred())
				Expect(director.DeployCalls()).To(Equal(1))
				Expect(director.ManifestBytes).To(Equal(deploymentManifestBytes))
			})

			It("returns the task ID", func() {
				Expect(returnedTaskID).To(Equal(taskID))
			})
		})
	})

})
