package boshclient

import (
	"encoding/json"
	"fmt"
	"io"
	"net/http"
	"time"
)

type BoshVMs map[string][]string

func (c *Client) VMs(name string) (BoshVMs, error) {
	errs := func(err error) (BoshVMs, error) {
		return nil, err
	}

	req, err := http.NewRequest("GET", fmt.Sprintf("%s/deployments/%s/vms?format=full", c.boshURL, name), nil)
	if err != nil {
		return errs(err)
	}

	authHeader, err := c.authHeaderBuilder.Build()
	if err != nil {
		return errs(err)
	}

	req.Header.Set("Authorization", authHeader)

	resp, err := c.HTTPClient.Do(req)
	if err != nil {
		return errs(err)
	}

	if resp.StatusCode != http.StatusFound {
		return errs(unexpectedStatusErr(resp, http.StatusFound))
	}

	taskID, err := taskIDFromRedirect(resp)
	if err != nil {
		return errs(err)
	}

	// TODO we only test the polling logic at system test level (when real bosh queues the task, and then taskes a long time to do it)
	// We should change that
	for {
		taskState, err := c.GetTaskState(taskID)
		if err != nil {
			return errs(err)
		}

		if taskState.State == Error {
			return nil, fmt.Errorf("task %d failed", taskID)
		}

		if taskState.State == Done {
			c.logger.Printf("Task %d finished with state %s\n", taskID, taskState)
			break
		}

		time.Sleep(time.Second * c.BoshPollingInterval)
	}

	vmsOutputForEachJob, err := c.VMsOutput(taskID)
	if err != nil {
		return errs(err)
	}

	vms := BoshVMs{}
	for _, vmsOutput := range vmsOutputForEachJob {
		vms[vmsOutput.InstanceGroup] = vmsOutput.IPs
	}

	return vms, nil
}

type BoshVMsOutput struct {
	IPs           []string
	InstanceGroup string `json:"job_name"`
}

func (c *Client) VMsOutput(taskID int) ([]BoshVMsOutput, error) {
	resp, err := c.getOutput(taskID)
	if err != nil {
		return nil, err
	}
	defer resp.Close()

	outputs := []BoshVMsOutput{}

	decoder := json.NewDecoder(resp)
	for decoder.More() {
		var output BoshVMsOutput
		if err := decoder.Decode(&output); err != nil {
			return nil, fmt.Errorf("error decoding bosh output response: %s", err)
		}
		outputs = append(outputs, output)
	}

	return outputs, nil
}

func (c *Client) getOutput(taskID int) (io.ReadCloser, error) {
	req, err := http.NewRequest("GET", fmt.Sprintf("%s/tasks/%d/output?type=result", c.boshURL, taskID), nil)
	if err != nil {
		return nil, fmt.Errorf("error creating request: %s", err)
	}

	authHeader, err := c.authHeaderBuilder.Build()
	if err != nil {
		return nil, err
	}

	req.Header.Set("Authorization", authHeader)

	resp, err := c.HTTPClient.Do(req)
	if err != nil {
		return nil, fmt.Errorf("error contacting bosh director: %s", err)
	}
	if resp.StatusCode != http.StatusOK {
		return nil, unexpectedStatusErr(resp, http.StatusOK)
	}
	return resp.Body, nil
}
