package broker

import (
	"fmt"
	"log"
	"time"

	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/serviceadapter"
)

//go:generate counterfeiter -o fake_bosh_client/fake_bosh_client.go . BoshClient
type BoshClient interface {
	DeployRaw(manifest []byte) (int, error)
	GetTaskState(taskID int) (boshclient.BoshTask, error)
	VMs(instanceID string) (boshclient.BoshVMs, error)
	DeleteDeployment(name string) (int, error)
	GetManifestRaw(name string) ([]byte, error)
}

type StaticBoshInfo struct {
	DirectorUUID    string
	StemcellOS      string
	StemcellVersion string
}

//go:generate counterfeiter -o fake_instance_repository/fake_instance_repository.go . InstanceRepository
type InstanceRepository interface {
	Save(i Instance) error
	FindByID(instanceID string) (Instance, error)
	AllInstances() ([]Instance, error)
	CountByPlanID(planID string) (int, error)
	DeleteByID(instanceID string) error
	InstanceExists(instanceID string) (bool, error)
}

type OperationType string

const (
	Create  = OperationType("Create")
	Update  = OperationType("Update")
	Upgrade = OperationType("Upgrade")
	Delete  = OperationType("Delete")
)

type Instance struct {
	ID                string
	PlanID            string
	BoshTaskID        *int
	LastOperationType OperationType
}

type SkippableError struct {
	error
}

func NewSkippableError(e error) error {
	return SkippableError{error: e}
}

//go:generate counterfeiter -o fake_deployer/fake_deployer.go . Deployer
type Deployer interface {
	Deploy(instance Instance, planID, deploymentName string, operationType OperationType, arbitraryParams map[string]interface{}, oldManifest []byte) (Instance, error)
}

//go:generate counterfeiter -o fake_service_adapter_client/fake_service_adapter_client.go . ServiceAdapterClient
type ServiceAdapterClient interface {
	GenerateManifest(boshInfo serviceadapter.BoshInfo, serviceReleases serviceadapter.ServiceReleases, plan serviceadapter.Plan, arbitraryParams map[string]interface{}, previousManifest []byte) ([]byte, error)
	CreateBinding(bindingID string, deploymentTopology boshclient.BoshVMs, manifest []byte) (map[string]interface{}, error)
	DeleteBinding(bindingID string, deploymentTopology boshclient.BoshVMs, manifest []byte) error
}

type Broker struct {
	boshClient     BoshClient
	instanceRepo   InstanceRepository
	serviceAdapter ServiceAdapterClient
	deployer       Deployer

	serviceOffering ServiceCatalog

	BoshPollingInterval time.Duration //TODO: Remove

	logger *log.Logger
}

func New(boshClient BoshClient, instanceRepo InstanceRepository, serviceAdapter ServiceAdapterClient,
	deployer Deployer, serviceOffering ServiceCatalog, logger *log.Logger) (*Broker, error) {

	existingPlanIDs, err := existingInstancePlanIDs(instanceRepo)
	if err != nil {
		return nil, err
	}

	for _, existingPlanID := range existingPlanIDs {
		if !containsPlanID(serviceOffering.Plans, existingPlanID) {
			return nil, fmt.Errorf("You cannot change the plan_id of a plan that has existing service instances")
		}
	}

	return &Broker{
		boshClient:     boshClient,
		instanceRepo:   instanceRepo,
		serviceAdapter: serviceAdapter,
		deployer:       deployer,

		serviceOffering: serviceOffering,

		BoshPollingInterval: 5,

		logger: logger,
	}, nil
}

func existingInstancePlanIDs(instanceRepo InstanceRepository) ([]string, error) {
	instances, err := instanceRepo.AllInstances()
	if err != nil {
		return nil, err
	}

	existingPlanIDs := []string{}
	for _, instance := range instances {
		existingPlanIDs = append(existingPlanIDs, instance.PlanID)
	}
	return existingPlanIDs, nil
}

func containsPlanID(plans []Plan, planID string) bool {
	for _, item := range plans {
		if item.ID == planID {
			return true
		}
	}
	return false
}
