package broker

import (
	"fmt"
	"log"

	"github.com/pivotal-cf/on-demand-service-broker/serviceadapter"
)

type deployer struct {
	boshClient      BoshClient
	instanceRepo    InstanceRepository
	serviceAdapter  ServiceAdapterClient
	serviceOffering ServiceCatalog
	staticBoshInfo  StaticBoshInfo
	serviceReleases serviceadapter.ServiceReleases
	logger          *log.Logger
}

func NewDeployer(
	boshClient BoshClient,
	instanceRepo InstanceRepository,
	serviceAdapter ServiceAdapterClient,
	serviceOffering ServiceCatalog,
	staticBoshInfo StaticBoshInfo,
	serviceReleases serviceadapter.ServiceReleases,
	logger *log.Logger,
) Deployer {
	return deployer{
		boshClient:      boshClient,
		instanceRepo:    instanceRepo,
		serviceAdapter:  serviceAdapter,
		serviceOffering: serviceOffering,
		staticBoshInfo:  staticBoshInfo,
		serviceReleases: serviceReleases,
		logger:          logger,
	}
}

func (d deployer) Deploy(instance Instance, planID, deploymentName string, operationType OperationType, arbitraryParams map[string]interface{}, oldManifest []byte) (Instance, error) {
	plan := d.serviceOffering.FindPlanById(planID)
	if plan == nil {
		return instance, SkippableError{fmt.Errorf("plan %s does not exist", instance.PlanID)}
	}

	boshInfo := serviceadapter.BoshInfo{
		Name:            deploymentName,
		DirectorUUID:    d.staticBoshInfo.DirectorUUID,
		StemcellOS:      d.staticBoshInfo.StemcellOS,
		StemcellVersion: d.staticBoshInfo.StemcellVersion,
	}

	d.logger.Printf("service adapter will generate manifest for instance %s\n", instance.ID)
	manifest, err := d.serviceAdapter.GenerateManifest(boshInfo, d.serviceReleases, serviceAdapterPlan(*plan), arbitraryParams, oldManifest)
	if err != nil {
		return instance, fmt.Errorf("error generating manifest: %s", err)
	}

	taskID, err := d.boshClient.DeployRaw(manifest)
	if err != nil {
		return instance, fmt.Errorf("error deploying instance: %s\n", err)
	}

	d.logger.Printf("Bosh task id for %s instance %s was %d\n", operationType, instance.ID, taskID)
	instance.BoshTaskID = &taskID
	instance.PlanID = planID
	instance.LastOperationType = operationType
	if err := d.instanceRepo.Save(instance); err != nil {
		return instance, fmt.Errorf("error saving instance: %s", err)
	}

	return instance, nil
}

func serviceAdapterPlan(plan Plan) serviceadapter.Plan {
	abridgedPlan := serviceadapter.Plan{
		Properties:     plan.Properties,
		InstanceGroups: plan.InstanceGroups,
	}
	return abridgedPlan
}
