package broker_test

import (
	"bytes"
	"errors"
	"fmt"
	"io"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/serviceadapter"
)

var _ = Describe("Deployer", func() {

	var (
		deployer          broker.Deployer
		instance          broker.Instance
		planID            string
		arbitraryParams   map[string]interface{}
		oldManifest       []byte
		deployedInstance  broker.Instance
		deployError       error
		operationType     broker.OperationType
		inMemoryLogBuffer *bytes.Buffer
		taskID            = 42

		manifest        = []byte("a-manifest")
		deploymentName  = "a-fancy-deployment"
		staticBoshInfo  = broker.StaticBoshInfo{DirectorUUID: "a-very-special-director-id"}
		serviceReleases = serviceadapter.ServiceReleases{
			{
				Name:    "name",
				Version: "vers",
				Jobs:    []string{"a", "b"},
			},
		}
	)

	BeforeEach(func() {
		inMemoryLogBuffer = bytes.NewBuffer([]byte{})
		combinedWriter := io.MultiWriter(inMemoryLogBuffer, GinkgoWriter)
		logger := log.New(combinedWriter, "[unit-tests] ", log.LstdFlags)
		deployer = broker.NewDeployer(boshClient, instanceRepo, serviceAdapter, serviceCatalog, staticBoshInfo, serviceReleases, logger)
		instance = broker.Instance{ID: "instance-id"}
		planID = existingPlanID
		operationType = broker.Create
	})

	JustBeforeEach(func() {
		deployedInstance, deployError = deployer.Deploy(instance, planID, deploymentName, operationType, arbitraryParams, oldManifest)
	})

	Context("deployment without previous manifest", func() {
		BeforeEach(func() {
			oldManifest = nil
		})
		Context("when bosh deploys the release successfully", func() {
			BeforeEach(func() {
				boshClient.DeployRawReturns(42, nil)
				instanceRepo.SaveReturns(nil)
				serviceAdapter.GenerateManifestReturns(manifest, nil)
			})

			It("generates a Redis manifest", func() {
				Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(1))
				actualBoshInfo, actualServiceReleases, actualPlan, actualArbParams, _ := serviceAdapter.GenerateManifestArgsForCall(0)
				Expect(actualBoshInfo).To(Equal(serviceadapter.BoshInfo{Name: deploymentName, DirectorUUID: staticBoshInfo.DirectorUUID}))
				Expect(actualArbParams).To(Equal(arbitraryParams))
				Expect(actualPlan).To(Equal(serviceadapter.Plan{InstanceGroups: existingPlan.InstanceGroups, Properties: existingPlan.Properties}))
				Expect(actualServiceReleases).To(Equal(serviceReleases))
			})

			It("Creates a bosh deployment using generated manifest", func() {
				Expect(boshClient.DeployRawCallCount()).To(Equal(1))
				deployedManifest := boshClient.DeployRawArgsForCall(0)
				Expect(deployedManifest).To(Equal(manifest))
			})

			It("saves the instance", func() {
				Expect(instanceRepo.SaveCallCount()).To(Equal(1))

				Expect(instanceRepo.SaveArgsForCall(0)).To(Equal(broker.Instance{
					ID:                "instance-id",
					PlanID:            existingPlanID,
					BoshTaskID:        &taskID,
					LastOperationType: operationType,
				}))
			})
		})
	})
	Context("logging", func() {
		BeforeEach(func() {
			boshClient.DeployRawReturns(42, nil)
			instanceRepo.SaveReturns(nil)
			serviceAdapter.GenerateManifestReturns(manifest, nil)
		})
		Context("for creating", func() {
			BeforeEach(func() {
				operationType = broker.Create
			})
			It("logs the bosh task ID returned by the director", func() {
				Expect(inMemoryLogBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task id for Create instance %s was %d", "instance-id", taskID)))
			})
		})
		Context("for updating", func() {
			BeforeEach(func() {
				operationType = broker.Update
			})
			It("logs the bosh task ID returned by the director", func() {
				Expect(inMemoryLogBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task id for Update instance %s was %d", "instance-id", taskID)))
			})
		})
		Context("for upgrading", func() {
			BeforeEach(func() {
				operationType = broker.Upgrade
			})
			It("logs the bosh task ID returned by the director", func() {
				Expect(inMemoryLogBuffer.String()).To(ContainSubstring(fmt.Sprintf("Bosh task id for Upgrade instance %s was %d", "instance-id", taskID)))
			})
		})
	})

	Context("deployment with previous manifest", func() {
		taskID := 42

		BeforeEach(func() {
			oldManifest = []byte("old-manifest-fetched-from-bosh")
			boshClient.DeployRawReturns(42, nil)
			serviceAdapter.GenerateManifestReturns(manifest, nil)
			arbitraryParams = map[string]interface{}{"foo": "bar"}
		})

		It("Creates a bosh deployment from the generated manifest", func() {
			Expect(boshClient.DeployRawCallCount()).To(Equal(1))
			deployedManifest := boshClient.DeployRawArgsForCall(0)
			Expect(deployedManifest).To(Equal(manifest))
		})

		It("generates manifest with correct inputs", func() {
			Expect(serviceAdapter.GenerateManifestCallCount()).To(Equal(1))
			boshInfoPassedToGenerate, actualServiceReleases, planPassedToGenerate, arbParamsPassedToGenerate, previousManifestPassedToGenerate := serviceAdapter.GenerateManifestArgsForCall(0)
			Expect(boshInfoPassedToGenerate).To(Equal(serviceadapter.BoshInfo{Name: deploymentName, DirectorUUID: staticBoshInfo.DirectorUUID}))
			Expect(planPassedToGenerate).To(Equal(serviceadapter.Plan{InstanceGroups: existingPlan.InstanceGroups, Properties: existingPlan.Properties}))
			Expect(arbParamsPassedToGenerate).To(Equal(arbitraryParams))
			Expect(previousManifestPassedToGenerate).To(Equal(oldManifest))
			Expect(actualServiceReleases).To(Equal(serviceReleases))
		})

		It("returns the deployed instace with the task id", func() {
			Expect(deployedInstance).To(Equal(broker.Instance{
				ID:                "instance-id",
				PlanID:            existingPlanID,
				BoshTaskID:        &taskID,
				LastOperationType: operationType,
			}))
		})
	})

	Context("when the manifest generator fails to generate the manifest", func() {
		BeforeEach(func() {
			serviceAdapter.GenerateManifestReturns(nil, errors.New("manifest fail"))
		})

		It("wraps the error", func() {
			Expect(deployError).To(MatchError(ContainSubstring("manifest fail")))
		})
	})
	Context("when the instance repository fails to save the task ID", func() {
		BeforeEach(func() {
			instanceRepo.SaveReturns(errors.New("failure"))
		})

		It("wraps the error", func() {
			Expect(deployError).To(MatchError(ContainSubstring("failure")))
		})
	})

	Context("when bosh fails to deploy the release", func() {
		BeforeEach(func() {
			boshClient.DeployRawReturns(0, errors.New("oops"))
		})

		It("wraps the error", func() {
			Expect(deployError).To(MatchError(ContainSubstring("oops")))
		})
	})

})
