package broker

import (
	"fmt"

	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
)

const (
	CreateInProgress = "Instance provisioning in progress"
	CreateCompleted  = "Instance provisioning completed"
	DeleteInProgress = "Instance deletion in progress"
	DeleteCompleted  = "Instance deletion completed"
	UpdateCompleted  = "Instance update completed"
	UpdateInProgress = "Instance update in progress"
)

var (
	taskStates = map[string]brokerapi.LastOperationState{
		boshclient.Queued:     brokerapi.InProgress,
		boshclient.Processing: brokerapi.InProgress,
		boshclient.Done:       brokerapi.Succeeded,
		boshclient.Error:      brokerapi.Failed,
	}

	descriptions = map[brokerapi.LastOperationState]map[OperationType]string{
		brokerapi.InProgress: {
			Create:  CreateInProgress,
			Update:  UpdateInProgress,
			Upgrade: "",
			Delete:  DeleteInProgress,
		},
		brokerapi.Succeeded: {
			Create:  CreateCompleted,
			Update:  UpdateCompleted,
			Upgrade: "",
			Delete:  DeleteCompleted,
		},
		brokerapi.Failed: {
			Create:  "Instance provisioning failed: %s",
			Update:  "Instance update failed: %s",
			Upgrade: "Failed for bosh task: %d",
			Delete:  "Instance deletion failed: %s",
		},
	}
)

func constructLastOperation(boshTask boshclient.BoshTask, instance Instance) brokerapi.LastOperation {
	taskState := taskStates[boshTask.State]
	template := descriptions[taskState][instance.LastOperationType]
	if boshTask.State == boshclient.Error {

		if instance.LastOperationType == Upgrade {
			template = fmt.Sprintf(template, *instance.BoshTaskID)
		} else {
			template = fmt.Sprintf(template, boshTask.Result)
		}
	}
	return brokerapi.LastOperation{State: taskState, Description: template}
}

func (b *Broker) LastOperation(instanceID string) (brokerapi.LastOperation, error) {
	errs := func(err error) (brokerapi.LastOperation, error) {
		b.logger.Println(err)
		return brokerapi.LastOperation{}, err
	}

	instance, err := b.instanceRepo.FindByID(instanceID)
	if err != nil {
		return errs(fmt.Errorf("error retrieving task ID for instance %s: %s", instanceID, err))
	}

	boshTask, err := b.boshClient.GetTaskState(*instance.BoshTaskID)
	if err != nil {
		return errs(fmt.Errorf("error retrieving task state from bosh: %s", err))
	}

	lastOperation := constructLastOperation(boshTask, instance)

	b.logLastOperation(instance, boshTask)

	if boshTask.State == boshclient.Error || (boshTask.State == boshclient.Done && boshTask.Description == boshclient.BoshTaskCreateDeployment) {
		if err := b.clearBoshTaskIDForInstance(instance); err != nil {
			return errs(fmt.Errorf("error saving instance %s: %s", instanceID, err))
		}
	}
	if boshTask.State == boshclient.Done && instance.LastOperationType == Delete {
		if err := b.instanceRepo.DeleteByID(instanceID); err != nil {
			return errs(fmt.Errorf("error deleting instance %s: %s", instanceID, err))
		}
	}
	return lastOperation, nil
}

func (b *Broker) logLastOperation(instance Instance, boshTask boshclient.BoshTask) {
	var state string
	switch boshTask.State {
	case boshclient.Done:
		state = "success"
	case boshclient.Error:
		state = "failure"
	default:
		return
	}

	var op string
	switch instance.LastOperationType {
	case Create:
		op = "creating"
	case Update:
		op = "updating"
	case Upgrade:
		op = "upgrading"
	case Delete:
		op = "deleting"
	}

	b.logger.Printf("task %d %s %s deployment for instance %s: %s\n", *instance.BoshTaskID, state, op, instance.ID, boshTask.Description)
}

func (b *Broker) clearBoshTaskIDForInstance(instance Instance) error {
	instance.BoshTaskID = nil
	return b.instanceRepo.Save(instance)
}
