package external_test

import (
	"encoding/json"
	"errors"
	"log"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/external"
	"github.com/pivotal-cf/on-demand-service-broker/external/fake_command_runner"
	"github.com/pivotal-cf/on-demand-service-broker/serviceadapter"
)

var _ = Describe("external service adapter", func() {
	const externalBinPath = "/thing"
	const manifestContent = "this is a genuine manifest"

	var (
		a         *external.Adapter
		cmdRunner *fake_command_runner.FakeCommandRunner
	)

	BeforeEach(func() {
		cmdRunner = new(fake_command_runner.FakeCommandRunner)
		a = &external.Adapter{
			CommandRunner:   cmdRunner,
			ExternalBinPath: externalBinPath,
			Logger:          log.New(GinkgoWriter, "[unit-tests] ", log.LstdFlags),
		}
	})

	Describe("generating manifest", func() {
		var (
			boshInfo         serviceadapter.BoshInfo
			serviceReleases  serviceadapter.ServiceReleases
			plan             serviceadapter.Plan
			params           map[string]interface{}
			previousManifest []byte

			manifest    []byte
			generateErr error
		)

		BeforeEach(func() {
			cmdRunner.RunReturns([]byte(manifestContent), []byte(""), nil)

			boshInfo = serviceadapter.BoshInfo{Name: "a-bosh-info"}
			serviceReleases = serviceadapter.ServiceReleases{
				{Name: "a-bosh-release"},
			}
			plan = serviceadapter.Plan{
				Properties: serviceadapter.Properties{
					"foo": "bar",
					"baz": map[interface{}]interface{}{
						"qux": "quux",
					},
				},
			}
			params = map[string]interface{}{
				"key": "value",
				"anotherkey": map[string]interface{}{
					"innerkey": "innervalue",
				},
			}
			previousManifest = []byte("a-manifest")
		})

		JustBeforeEach(func() {
			manifest, generateErr = a.GenerateManifest(boshInfo, serviceReleases, plan, params, previousManifest)
		})

		It("invokes external manifest generator with serialised parameters", func() {
			serialisedBoshInfo, err := json.Marshal(boshInfo)
			Expect(err).NotTo(HaveOccurred())

			serialisedServiceReleases, err := json.Marshal(serviceReleases)
			Expect(err).NotTo(HaveOccurred())

			plan.Properties = external.SanitiseForJSON(plan.Properties)
			serialisedPlan, err := json.Marshal(plan)
			Expect(err).NotTo(HaveOccurred())

			serialisedParams, err := json.Marshal(params)
			Expect(err).NotTo(HaveOccurred())

			Expect(cmdRunner.RunCallCount()).To(Equal(1))
			argsPassed := cmdRunner.RunArgsForCall(0).Args
			Expect(argsPassed).To(ConsistOf(externalBinPath, "generate-manifest", string(serialisedBoshInfo),
				string(serialisedServiceReleases), string(serialisedPlan),
				string(serialisedParams), string(previousManifest)))
		})

		Context("when the external service adapter succeeds", func() {
			It("returns no error", func() {
				Expect(generateErr).ToNot(HaveOccurred())
			})

			It("returns the deserialised stdout from the external process as a bosh manifest", func() {
				Expect(manifest).To(Equal([]byte(manifestContent)))
			})
		})

		Context("when the external service adapter fails", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), errors.New("oops"))
			})

			It("returns an error", func() {
				Expect(generateErr).To(MatchError("an error occurred running external service adapter at /thing: 'oops'. Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
			})
		})
	})

	Describe("creating binding", func() {

		var (
			bindingID          string
			deploymentTopology boshclient.BoshVMs
			manifest           []byte

			credentials      map[string]interface{}
			createBindingErr error
		)

		BeforeEach(func() {
			cmdRunner.RunReturns([]byte(`{"username":"bob"}`), []byte(""), nil)

			bindingID = "the-binding"
			deploymentTopology = boshclient.BoshVMs{"the-deployment": []string{"a-vm"}}
			manifest = []byte("a-manifest")
		})

		JustBeforeEach(func() {
			credentials, createBindingErr = a.CreateBinding(bindingID, deploymentTopology, manifest)
		})

		It("invokes external binding creator with serialised params", func() {
			serialisedVMs, err := json.Marshal(deploymentTopology)
			Expect(err).NotTo(HaveOccurred())

			Expect(cmdRunner.RunCallCount()).To(Equal(1))
			argsPassed := cmdRunner.RunArgsForCall(0).Args
			Expect(argsPassed).To(ConsistOf(externalBinPath, "create-binding", bindingID, string(serialisedVMs), string(manifest)))
		})

		Context("when the external adapter succeeds", func() {
			It("returns no error", func() {
				Expect(createBindingErr).ToNot(HaveOccurred())
			})

			It("returns the service-specific binding output", func() {
				Expect(credentials).To(Equal(map[string]interface{}{"username": "bob"}))
			})
		})

		Context("when the external service adapter fails", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), errors.New("oops"))
			})

			It("returns an error", func() {
				Expect(createBindingErr).To(MatchError("an error occurred running external service adapter at /thing: 'oops'. Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
			})
		})
	})

	Describe("deleting binding", func() {
		var (
			bindingID          string
			deploymentTopology boshclient.BoshVMs
			manifest           []byte

			deleteBindingError error
		)

		BeforeEach(func() {
			cmdRunner.RunReturns([]byte(""), []byte(""), nil)

			bindingID = "the-binding"
			deploymentTopology = boshclient.BoshVMs{}
			manifest = []byte("a-manifest")
		})

		JustBeforeEach(func() {
			deleteBindingError = a.DeleteBinding(bindingID, deploymentTopology, manifest)
		})

		It("invokes external binary with params to delete binding", func() {
			serialisedBoshVMs, err := json.Marshal(deploymentTopology)
			Expect(err).NotTo(HaveOccurred())

			Expect(cmdRunner.RunCallCount()).To(Equal(1))
			argsPassed := cmdRunner.RunArgsForCall(0).Args
			Expect(argsPassed).To(ConsistOf(externalBinPath, "delete-binding", bindingID, string(serialisedBoshVMs), string(manifest)))
		})

		Context("when the external adapter succeeds", func() {
			It("returns no error", func() {
				Expect(deleteBindingError).ToNot(HaveOccurred())
			})
		})

		Context("when the external adapter fails", func() {
			BeforeEach(func() {
				cmdRunner.RunReturns([]byte("I'm stdout"), []byte("I'm stderr"), errors.New("oops"))
			})

			It("returns an error", func() {
				Expect(deleteBindingError).To(MatchError("an error occurred running external service adapter at /thing: 'oops'. Stdout: 'I'm stdout', stderr: 'I'm stderr'"))
			})
		})
	})
})
