package upgrader

import (
	"encoding/json"
	"fmt"
	"net/http"
	"time"

	"github.com/pivotal-cf/brokerapi"
)

func Upgrade(brokerUsername, brokerPassword, brokerUrl string, pollingInterval int64) ([]string, []string, error) {
	errs := func(err error) ([]string, []string, error) {
		return nil, nil, err
	}

	instances, err := getServiceInstances(brokerUsername, brokerPassword, brokerUrl)
	if err != nil {
		return errs(fmt.Errorf("error getting service instances: %s\n", err))
	}

	skippedInstances := []string{}
	upgradedInstances := []string{}

	for _, instance := range instances {
		upgradeStarted, err := upgradeInstance(brokerUsername, brokerPassword, brokerUrl, instance)
		if err != nil {
			return errs(fmt.Errorf("Upgrade failed for service_instance_id %s: %s\n", instance, err))
		}

		if !upgradeStarted {
			skippedInstances = append(skippedInstances, instance)
			continue
		}

		lastOperation := brokerapi.LastOperation{State: brokerapi.InProgress}
		for lastOperation.State == brokerapi.InProgress {
			time.Sleep(time.Millisecond * time.Duration(pollingInterval))
			lastOperation, err = getLastOperation(brokerUsername, brokerPassword, brokerUrl, instance)
			if err != nil {
				return errs(fmt.Errorf("error getting last operation: %s\n", err))
			}
		}

		if lastOperation.State != brokerapi.Succeeded {
			return errs(fmt.Errorf("Upgrade failed for service_instance_id %s: %s", instance, lastOperation.Description))
		}
		upgradedInstances = append(upgradedInstances, instance)
	}
	return upgradedInstances, skippedInstances, nil
}

func getServiceInstances(brokerUsername, brokerPassword, brokerUrl string) ([]string, error) {
	var instances []string
	if err := makeRequest(brokerUsername, brokerPassword, brokerUrl, "/mgmt/service_instances", &instances); err != nil {
		return nil, err
	}
	return instances, nil
}

func upgradeInstance(brokerUsername, brokerPassword, brokerUrl, instance string) (bool, error) {

	request, err := http.NewRequest("PATCH", fmt.Sprintf("%s%s", brokerUrl, "/mgmt/service_instances/"+instance), nil)
	if err != nil {
		return false, err
	}
	request.SetBasicAuth(brokerUsername, brokerPassword)

	response, err := http.DefaultClient.Do(request)
	if err != nil {
		return false, err
	}
	defer response.Body.Close()

	switch response.StatusCode {
	case http.StatusAccepted:
		return true, nil
	case http.StatusConflict:
		return false, nil
	default:
		var errorResponse brokerapi.ErrorResponse
		if err := json.NewDecoder(response.Body).Decode(&errorResponse); err != nil {
			return false, fmt.Errorf("cant parse response %s", err)
		}
		return false, fmt.Errorf("ERROR: %s", errorResponse.Description)
	}
}

func getLastOperation(brokerUsername, brokerPassword, brokerUrl, instance string) (brokerapi.LastOperation, error) {
	var lastOperation brokerapi.LastOperation
	if err := makeRequest(brokerUsername, brokerPassword, brokerUrl, "/v2/service_instances/"+instance+"/last_operation", &lastOperation); err != nil {
		return brokerapi.LastOperation{}, err
	}
	return lastOperation, nil
}

func makeRequest(brokerUsername, brokerPassword, brokerUrl, requestPath string, responseData interface{}) error {
	request, err := http.NewRequest("GET", fmt.Sprintf("%s%s", brokerUrl, requestPath), nil)
	if err != nil {
		return err
	}

	request.SetBasicAuth(brokerUsername, brokerPassword)

	response, err := http.DefaultClient.Do(request)
	if err != nil {
		return err
	}

	if response.StatusCode != 200 {
		return fmt.Errorf("response returned bad status: %s", response.Status)
	}

	err = json.NewDecoder(response.Body).Decode(responseData)
	if err != nil {
		return err
	}

	return nil
}
